(*
 * <COPYRIGHT>
 * Copyright 2005 Gerd Stolpmann
 *
 * <GPL>
 * This file is part of UMLMON.
 *
 * UMLMON is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * UMLMON is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with UMLMON; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * </>
 *)

(* $Id: umldir.ml 188 2005-10-14 11:40:29Z gerd $ *)

(* Directory service *)

open Umlmon_types
open Umlmon_common
open Umlmonctrl_aux
open Printf

module Client = Umlmonctrl_clnt.Ctrl.V1


let proc_ping session () reply =
  reply ()
;;


let umlmon_client_path global id =
  Filename.concat 
    (Filename.concat global.umlvardir id) "ctrl"
;;


let umlmon_client ?(ues = Unixqueue.create_unix_event_system()) global id =
  let path = umlmon_client_path global id in
  let client =
    Client.create_client
      ~esys:ues (Rpc_client.Unix path) Rpc.Tcp in
  Rpc_client.configure client 0 5.0;   (* Wait at most 5 seconds *)
  client
;;


let umlmon_client_descr ?(ues = Unixqueue.create_unix_event_system()) d =
  let client =
    Client.create_client
      ~esys:ues (Rpc_client.Descriptor d) Rpc.Tcp in
  Rpc_client.configure client 0 5.0;   (* Wait at most 5 seconds *)
  client
;;


let proc_get_monitors global ues session () reply =
  (* TODO: Query monitors in parallel *)
  try
    let list = Umlmon_config.vm_list() in
    let list' =
      (List.flatten
	 (List.map
	    (fun id ->
	       try
		 let client = umlmon_client global id in
		 try
		   let r =
		     { monitor_id = id; monitor_port = Client.port client () } in
		   Rpc_client.shut_down client;
		   [ r ]
		 with
		     _ ->
		       Rpc_client.shut_down client;
		       []    (* Hide all errors! *)
	       with
		   _ -> []   (* Hide all errors! *)
	    )
	    list)) in
    reply (Array.of_list list')
  with
      err ->
	Rpc_server.reply_error session Rpc.System_err;
	raise err
;;


let get_port global id =
  try
    let client = umlmon_client global id in
    try
      let port_opt = Client.port client () in
      Rpc_client.shut_down client;
      port_opt
    with
	_ ->
	   Rpc_client.shut_down client;
	  None
  with
      _ ->
	None
;;


let proc_get_port global ues session id reply =
  reply (get_port global id)
;;


let proc_have_redirection global ues session () reply =
  reply global.umltcp_redir
;;


let create_dir_server_for global connector proto mode ues =
  Umlmonctrl_srv.Directory.V1.create_async_server
    ~proc_ping
    ~proc_get_monitors:(proc_get_monitors global ues)
    ~proc_get_port:(proc_get_port global ues)
    ~proc_have_redirection:(proc_have_redirection global ues)
    connector
    proto
    mode
    ues
;;


let create_dir_server global ues =
  (* Create Unix Domain server: *)
  if not (Sys.file_exists global.umlvardir) then (
    Unix.mkdir global.umlvardir 0o755
  );
  let sock_dir = Filename.concat global.umlvardir "directory" in
  if not (Sys.file_exists sock_dir) then (
    Unix.mkdir sock_dir 0o755
  );
  let () =
    let sockname = 
      Filename.concat sock_dir "directory" in
    (try Sys.remove sockname with _ -> ());
    let srv =
      create_dir_server_for
	global
	(Rpc_server.Unix sockname)
	Rpc.Tcp
	Rpc.Socket
	ues in
    Unix.chmod sockname 0o666;
    ()
  in
  (* Create TCP server: *)
  let () =
    let serv = 
      try Unix.getservbyname "umldir" "tcp"
      with Not_found -> failwith "umldir/tcp service not found (probably in /etc/services)" in
    if global.umltcp || global.umltcp_redir then (
      let srv =
	create_dir_server_for
	  global
	  (Rpc_server.Internet(Unix.inet_addr_any,serv.Unix.s_port))
	  Rpc.Tcp
	  Rpc.Socket
	  ues in
      Rpc_server.set_session_filter srv (fun _ -> `Accept_limit_length(256, `Deny));
      ()
    )
  in
  ()
;;


class async_string_channel s =
object(self)
  inherit Netchannels.input_string s

  method can_input = true
  method request_notification (_ : unit->bool) = ()
end


let establish_redir_conn global ues s =
  let close_s = ref false in
  (* whether to close [s] from the RPC server's onclose action *)

  let proc_ensure_id session id =
    (* Detach the RPC server from [s]: *)
    close_s := true; 
    Rpc_server.stop_connection 
      (Rpc_server.get_server session)
      (Rpc_server.get_connection_id session);
    (* Note: We detach early, because we cannot use an Rpc_server for the
     * reply. The Rpc_server does not indicate when the reply is fully
     * sent, so we would not know when to switch to copy mode. The 
     * workaround is to create the reply manually and to send it using
     * a raw Uq_engines.sender object.
     *)
    (* The user wants to connect to the monitor [id], so connect: *)
    let path = umlmon_client_path global id in
    let monitor_s =
      Unix.socket Unix.PF_UNIX Unix.SOCK_STREAM 0 in
    ( try
	Umlmon_common.restart
	  (Unix.connect monitor_s) (Unix.ADDR_UNIX path);
	(* Our first action is to reset the authenticator, so the user _must_
	 * authenticate. For simplicity, this is a synchronous call.
	 *)
	let monitor_client = umlmon_client_descr monitor_s in
	( try
	    Client.auth_reset monitor_client ();
             (* Everything ok up to now, so we can send a reply: *)
             let reply_pv = 
	       Rpc_packer.pack_successful_reply
		 Umlmonctrl_aux.program_Ctrl'V1
		 "ensure_id"
		 (Rpc_server.get_xid session)
		 "AUTH_NONE"
		 ""
		 Xdr.XV_void in
	     let reply_str = Rpc_packer.rm_string_of_packed_value reply_pv in
	     let l = String.length reply_str - 4 in   (* Set record mark *)
             let l_uint4 = Rtypes.uint4_of_int l in
             let l_uint4_32 = Rtypes.logical_int32_of_uint4 l_uint4 in
             let l_uint4_32' = Int32.logor l_uint4_32 Int32.min_int in
             (* 0x80000000 *)
             let l_uint4' = Rtypes.logical_uint4_of_int32 l_uint4_32' in
             Rtypes.write_uint4 reply_str 0 l_uint4';
	     close_s := false;   (* because the copier will close [s] *)
	     (* Send the reply using a sender engine: *)
	     let reply_ch = new async_string_channel reply_str in
	     let sender = 
	       new Uq_engines.sender 
		 ~src:reply_ch ~dst:s ~close_dst:false ues in
	     (* After that: set up the copier: *)
	     Uq_engines.when_state
	       ~is_done:(fun () ->
			   let copier = 
			     new Uq_engines.copier (`Bidirectional(s,monitor_s)) ues in
			   (* The copier does the rest in an autonomous way! *)
			   ())
	       ~is_error:(fun err -> Unix.close s; Unix.close monitor_s; raise err)
	       sender
	  with
	    | err -> 
		Rpc_client.shut_down monitor_client; raise err
	)
      with
	| Unix.Unix_error(Unix.ECONNREFUSED,_,_) -> Unix.close monitor_s
	| err -> Unix.close monitor_s; raise err
    )
  in

  let srv =
    Rpc_server.create 
      ues
      (Rpc_server.Descriptor s)
      Rpc.Tcp
      Rpc.BiPipe
      Umlmonctrl_aux.program_Ctrl'V1
      [ Rpc_server.Async 
	  { Rpc_server.async_name = "ensure_id";
	    Rpc_server.async_invoke = 
	      (fun session arg ->
		 proc_ensure_id 
		   session 
		   (Umlmonctrl_aux._to_Ctrl'V1'ensure_id'arg arg))
	  }
      ]
      20 (* backlog limit *) in
  Rpc_server.set_session_filter srv (fun _ -> `Accept_limit_length(256, `Deny));
  Rpc_server.set_onclose_action srv (fun _ -> 
				       if !close_s then Unix.close s);
  ()
;;


let rec accept_redir_conn global ues acceptor =
  let acc_eng = acceptor # accept() in
  Uq_engines.when_state
    ~is_done:(fun (s, _) ->
		accept_redir_conn global ues acceptor;
		establish_redir_conn global ues s
	     )
    ~is_error:(fun exn -> raise exn)
    acc_eng
;;


let redir_options =
  { Uq_engines.lstn_backlog = 20;
    lstn_reuseaddr = true
  }
;;


let create_redir_server global ues =
  let () =
    let serv = 
      try Unix.getservbyname "umlredir" "tcp"
      with Not_found -> failwith "umlredir/tcp service not found (probably in /etc/services)" in
    if global.umltcp_redir then (
      let lstn_eng =
	Uq_engines.listener (`Socket(`Sock_inet(Unix.SOCK_STREAM,
						Unix.inet_addr_any,
						serv.Unix.s_port),
				     redir_options)) ues in
      Uq_engines.when_state
	~is_done:(fun acceptor -> 
		    accept_redir_conn global ues acceptor
		 )
	~is_error:(fun exn -> raise exn)
	lstn_eng
    )
  in
  ()
;;



let version() =
  printf "UMLMON Release %s Revision %s\n" Instconf.release Instconf.revision;
  printf "Copyright 2005 by Informatikbuero Gerd Stolpmann.\n\n";
  printf "UMLMON is free software, and can be distributed under the terms\n";
  printf "of the GNU General Public License.\n\n";
  printf "Commercial support available, please contact gerd@gerd-stolpmann.de\n";
  flush stdout;
  exit 0
;;


let main() =
  let fg = ref false in

  Arg.parse
    (Arg.align
       [ "-fg", Arg.Set fg, " Run in the foreground";
	"-version", Arg.Unit version, " Print version number and exit";
       ])
    (fun s ->
       raise(Arg.Bad "Unexpected argument"))
    "usage: umldir [options]";
  
  if not (Sys.file_exists "/var/run/umlmon") then (
    Unix.mkdir "/var/run/umlmon" 0o755
  );

  let ues = Unixqueue.create_unix_event_system() in
  let global = Umlmon_config.read_global_config() in
  create_dir_server global ues;
  create_redir_server global ues;
  daemon ~wait:!fg
    "/var/run/umlmon/directory.pid"
    (fun () ->
       let cont = ref true in
       while !cont do
	 try
	   Unixqueue.run ues;
	   cont := false
	 with
	   | Sys.Break -> cont := false
	   | err ->
	       prerr_endline ("umldir: Caught exception: " ^ Printexc.to_string err);
       done
    )
;;


try
  Sys.set_signal Sys.sigpipe Sys.Signal_ignore;
  ignore(Unix.umask 0o22);
  main();
with
  | Failure msg ->
      prerr_endline ("umldir: " ^ msg); exit 1
  | Sys_error msg ->
      prerr_endline ("umldir: " ^ msg); exit 1
  | Unix.Unix_error(_,_,_) as e ->
      prerr_endline ("umldir: " ^ Umlmon_common.string_of_unix_error e); exit 1
  | err ->
      prerr_endline ("umldir: Caught exception " ^ Printexc.to_string err); exit 1
;;
