(*
 * <COPYRIGHT>
 * Copyright 2005 Gerd Stolpmann
 *
 * <GPL>
 * This file is part of UMLMON.
 *
 * UMLMON is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * UMLMON is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with UMLMON; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * </>
 *)

(* $Id: umlmon_net.ml 196 2005-10-16 14:52:26Z gerd $ *)

open Umlmon_types
open Umlmon_common
open Printf

let do_modprobe_tun () =
  (* This is for udev-based systems *)
  if not (Sys.file_exists "/dev/net/tun") then (
    monitor_log "No /dev/net/tun found. Trying to 'modprobe tun'";
    ignore(Sys.command "modprobe tun");
    let n = ref 0 in
    while not (Sys.file_exists "/dev/net/tun") do
      Unix.sleep 1;
      incr n;
      if !n > 5 then failwith "No TUN/TAP driver";
    done
  )
;;


type net_setup = 
    { net_kernel_args : string list }
;;


let setup_host_to_host gs eth hth =
  let global = Lazy.force gs.umlglobal in
  let spec = Lazy.force gs.umlspec in
  let uid = spec.umlpriv.priv_uid in
  let gid = spec.umlpriv.priv_gid in
  let err_actions = ref [ fun () -> 
			    Umlmon_shared.release_all_hwaddr global spec ] in
  ( try
      monitor_log ("Setting up network for " ^ eth.ethname);
      (* Setup TAP device: *)
      let tun_fd = Unix.openfile "/dev/net/tun" [Unix.O_RDWR] 0 in
      err_actions := (fun () -> Unix.close tun_fd) :: !err_actions;
      let tap = 
	if Sys.file_exists
	  (sprintf "/proc/sys/net/ipv4/conf/%s" hth.hth_proxy_if) 
	then
	  hth.hth_proxy_if
	else (
	  let tap =
	    Linux.Tun.create_tap tun_fd hth.hth_proxy_if uid in
	  err_actions := 
	    (fun () -> Linux.Tun.delete_tap tun_fd tap) :: !err_actions;
	  if tap <> hth.hth_proxy_if then
	    failwith "Bad TAP device name";
	  tap
	) in
      (* Maybe we should set the HW address: *)
      let conf_hwaddr =
	match hth.hth_proxy_hwaddr with
	  | `Unconfigured -> ""
	  | `HW hwaddr ->
	      format_hwaddr hwaddr
	  | `HW_from_pool ->
	      let hwaddr = 
		Umlmon_shared.reserve_hwaddr global spec (Some hth.hth_proxy_addr) in
	      format_hwaddr hwaddr
      in
      (* Configure TAP device: *)
      (* Note: The device must be down to set the hw address. So ensure this
       * first.
       *)
      ignore(Sys.command(sprintf "ifconfig %s down" tap));
      if conf_hwaddr <> "" then
	ignore(Sys.command(sprintf "ifconfig %s hw ether %s" tap conf_hwaddr));
      ignore(Sys.command(sprintf "ifconfig %s %s netmask 255.255.255.255 up" tap hth.hth_proxy_addr));
      ignore(Sys.command(sprintf "route add -host %s dev %s"
			   hth.hth_guest_addr tap));
      ( match hth.hth_host_if with
	  | Some d ->
	      let proc_f = 
		open_out (sprintf "/proc/sys/net/ipv4/conf/%s/proxy_arp" tap) in
	      err_actions := (fun () -> close_out proc_f) :: !err_actions;
	      output_string proc_f "1\n";
	      close_out proc_f;
	      ignore(Sys.command(sprintf "arp -Ds %s %s pub" 
				   hth.hth_guest_addr d));
	  | None -> ()
      );
      Unix.close tun_fd;
      (* Determine kernel arguments: *)
      let conf_hwaddr =
	match hth.hth_guest_hwaddr with
	  | `Unconfigured -> ""
	  | `HW hwaddr -> "," ^ format_hwaddr hwaddr
	  | `HW_from_pool ->
	      let hwaddr = 
		Umlmon_shared.reserve_hwaddr global spec (Some hth.hth_guest_addr) in
	       "," ^ format_hwaddr hwaddr
      in
      [ eth.ethname ^ "=tuntap," ^ hth.hth_proxy_if ^ conf_hwaddr ]

    with
	err ->
	  List.iter (fun f -> f()) !err_actions;
	  raise err
  )
;;


let setup_bridged gs eth br =
  let global = Lazy.force gs.umlglobal in
  let spec = Lazy.force gs.umlspec in
  let uid = spec.umlpriv.priv_uid in
  let gid = spec.umlpriv.priv_gid in
  let err_actions = ref [ fun () -> 
			    Umlmon_shared.release_all_hwaddr global spec ] in
  ( try
      monitor_log ("Setting up network for " ^ eth.ethname);
      (* Setup TAP device: *)
      let tun_fd = Unix.openfile "/dev/net/tun" [Unix.O_RDWR] 0 in
      err_actions := (fun () -> Unix.close tun_fd) :: !err_actions;
      let tap = 
	if Sys.file_exists
	  (sprintf "/proc/sys/net/ipv4/conf/%s" br.br_proxy_if) 
	then
	  br.br_proxy_if
	else (
	  let tap =
	    Linux.Tun.create_tap tun_fd br.br_proxy_if uid in
	  err_actions := 
	    (fun () -> Linux.Tun.delete_tap tun_fd tap) :: !err_actions;
	  if tap <> br.br_proxy_if then
	    failwith "Bad TAP device name";
	  tap
	) in
      (* Maybe we should set the HW address: *)
      let conf_hwaddr =
	match br.br_proxy_hwaddr with
	  | `Unconfigured -> ""
	  | `HW hwaddr ->
	      format_hwaddr hwaddr
	  | `HW_from_pool ->
	      let hwaddr = 
		Umlmon_shared.reserve_hwaddr global spec None in
	      format_hwaddr hwaddr
      in
      (* Configure TAP device: *)
      (* Note: The device must be down to set the hw address. So ensure this
       * first.
       *)
      ignore(Sys.command(sprintf "ifconfig %s down" tap));
      if conf_hwaddr <> "" then
	ignore(Sys.command(sprintf "ifconfig %s hw ether %s" tap conf_hwaddr));
      ignore(Sys.command(sprintf "ifconfig %s up" tap));
      (* Add the TAP device to the bridge *)
      ignore(Sys.command(sprintf "brctl addif %s %s" br.br_bridge_if tap));
      Unix.close tun_fd;
      (* Determine kernel arguments: *)
      let conf_hwaddr =
	match br.br_guest_hwaddr with
	  | `Unconfigured -> ""
	  | `HW hwaddr -> "," ^ format_hwaddr hwaddr
	  | `HW_from_pool ->
	      let hwaddr = 
		Umlmon_shared.reserve_hwaddr global spec None in
	       "," ^ format_hwaddr hwaddr
      in
      [ eth.ethname ^ "=tuntap," ^ br.br_proxy_if ^ conf_hwaddr ]

    with
	err ->
	  List.iter (fun f -> f()) !err_actions;
	  raise err
  )
;;


let setup_networking gs =
  let spec = Lazy.force gs.umlspec in
  let uid = spec.umlpriv.priv_uid in
  let gid = spec.umlpriv.priv_gid in
  let have_tun =
    List.exists 
      (fun eth -> match eth.ethdef with | `Host_to_host _ -> true | `Bridged _ -> true | _ -> false)
      spec.umleths in
  if have_tun then
    do_modprobe_tun();
  let kernel_args = 
    List.flatten
      (List.map
	 (fun eth ->
	    match eth.ethdef with
	      | `Host_to_host hth ->
		  setup_host_to_host gs eth hth
	      | `Bridged br ->
		  setup_bridged gs eth br
	      | `Unparsed s ->
		  [ eth.ethname ^ "=" ^ s ]
	 )
	 spec.umleths
      ) in
  { net_kernel_args = kernel_args }
;;


let clean_networking gs =
  let global = Lazy.force gs.umlglobal in
  let spec = Lazy.force gs.umlspec in
  List.iter
    (fun eth ->
       match eth.ethdef with
	 | `Host_to_host hth ->
	     monitor_log ("Shutting down network for " ^ eth.ethname);
	     ( match hth.hth_host_if with
		 | Some d ->
		     ignore(Sys.command(sprintf "arp -i %s -d %s pub" d hth.hth_guest_addr));
		 | None -> ()
	     );
	     let tun_fd = Unix.openfile "/dev/net/tun" [Unix.O_RDWR] 0 in
	     Linux.Tun.delete_tap tun_fd hth.hth_proxy_if;
	     Unix.close tun_fd;
	 | `Bridged br ->
	     monitor_log ("Shutting down network for " ^ eth.ethname);
	     ignore(Sys.command(sprintf "brctl delif %s %s" br.br_bridge_if br.br_proxy_if));
	     let tun_fd = Unix.openfile "/dev/net/tun" [Unix.O_RDWR] 0 in
	     Linux.Tun.delete_tap tun_fd br.br_proxy_if;
	     Unix.close tun_fd;
	 | `Unparsed _ ->
	     ()
    )
    spec.umleths;
  Umlmon_shared.release_all_hwaddr global spec;
;;


let kernel_arguments setup = setup.net_kernel_args ;;
